#include "z64.h"
#include "Gfx #1.3.h"
#include "vi.h"
#include "rdp.h"
#include "bitmap.h"

static const int screen_width = SCREEN_WIDTH, screen_height = SCREEN_HEIGHT;

LPDIRECTDRAW7 lpdd = 0;
LPDIRECTDRAWSURFACE7 lpddsprimary; 
LPDIRECTDRAWSURFACE7 lpddsback;
DDSURFACEDESC2 ddsd;
HRESULT res;
RECT dst, src;
INT32 pitchindwords;

FILE* zeldainfo = 0;
int ProcessDListShown = 0;
extern int SaveLoaded;
extern UINT32 command_counter;

GFX_INFO gfx;

static char filepath[_MAX_PATH];

EXPORT void CALL CaptureScreen ( char * Directory )
{
    static unsigned short count;
    register int i;

    for (i = 0; i < _MAX_PATH; i++)
        if (Directory[i] == '\0')
            break;
        else
            filepath[i] = *(Directory + i);
    if (i + 12 >= _MAX_PATH)
    {
        MessageBox(NULL, "Screenshots directory too long.", NULL, MB_ICONERROR);
        return;
    }
    filepath[i] = GET_GFX_INFO(HEADER)[0x3B ^ BYTE_ADDR_XOR];
    filepath[i++] = (filepath[i] == '\0') ? ' ' : filepath[i];
    filepath[i] = GET_GFX_INFO(HEADER)[0x3C ^ BYTE_ADDR_XOR];
    filepath[i++] = (filepath[i] == '\0') ? ' ' : filepath[i];
    filepath[i] = GET_GFX_INFO(HEADER)[0x3D ^ BYTE_ADDR_XOR];
    filepath[i++] = (filepath[i] == '\0') ? ' ' : filepath[i];
    filepath[i] = GET_GFX_INFO(HEADER)[0x3E ^ BYTE_ADDR_XOR];
    filepath[i++] = (filepath[i] == '\0') ? ' ' : filepath[i];
    filepath[i++] = '0' | (count/1000 % 10);
    filepath[i++] = '0' | (count/ 100 % 10);
    filepath[i++] = '0' | (count/  10 % 10);
    filepath[i++] = '0' | (count/   1 % 10);
    filepath[i++] = '.';
    filepath[i++] = 'b';
    filepath[i++] = 'm';
    filepath[i++] = 'p';
    filepath[i] = '\0';
    capture_screen_to_file[*GET_GFX_INFO(VI_STATUS_REG) & 0x00000001](filepath);
    count = (count + 0x0001) % 10000;
    if (count == 9999)
    {
        MessageBox(
            NULL, "Screenshots directory is about to overflow.", "Warning",
            MB_ICONWARNING);
    }
    return;
}

EXPORT void CALL ChangeWindow (void)
{
}

EXPORT void CALL CloseDLL (void)
{
    return;
}

EXPORT void CALL ReadScreen(void **dest, long *width, long *height)
{
}

 
EXPORT void CALL DrawScreen (void)
{
}

EXPORT void CALL GetDllInfo(PLUGIN_INFO* PluginInfo)
{
    PluginInfo -> Version = 0x0103;
    PluginInfo -> Type  = PLUGIN_TYPE_GFX;
    strcpy(
#if (SCREEN_WIDTH == 320 && SCREEN_HEIGHT == 240)
    PluginInfo -> Name, "angrylion's RDP (320x240)"
#else
    PluginInfo -> Name, "angrylion's RDP"
#endif
    );
    PluginInfo -> NormalMemory = TRUE;
    PluginInfo -> MemoryBswaped = TRUE;
    return;
}

EXPORT BOOL CALL InitiateGFX (GFX_INFO Gfx_Info)
{
  gfx = Gfx_Info;
  
  return TRUE;
}

 
EXPORT void CALL MoveScreen (int xpos, int ypos)
{
    RECT statusrect;
    POINT p;
    p.x = p.y = 0;
    GetClientRect(gfx.hWnd, &dst);
    ClientToScreen(gfx.hWnd, &p);
    OffsetRect(&dst, p.x, p.y);
    GetClientRect(gfx.hStatusBar, &statusrect);
    dst.bottom -= statusrect.bottom;

    
    
    
    
    
    
    
}

 
EXPORT void CALL ProcessDList(void)
{
    if (!ProcessDListShown)
    {
        DisplayError("ProcessDList");
        ProcessDListShown = 1;
    }
}

EXPORT void CALL ProcessRDPList(void)
{
    process_RDP_list();
    return;
}

EXPORT void CALL RomClosed (void)
{
    rdp_close();
    if (lpddsback)
    {
        IDirectDrawSurface_Release(lpddsback);
        lpddsback = 0;
    }
    if (lpddsprimary)
    {
        IDirectDrawSurface_Release(lpddsprimary);
        lpddsprimary = 0;
    }
    if (lpdd)
    {
        IDirectDraw_Release(lpdd);
        lpdd = 0;
    }

    SaveLoaded = 1;
    command_counter = 0;
}

 
EXPORT void CALL RomOpen (void)
{
    DDPIXELFORMAT ftpixel;
    LPDIRECTDRAWCLIPPER lpddcl;
    RECT bigrect, smallrect, statusrect;
    POINT p;
    int rightdiff;
    int bottomdiff;

    GetWindowRect(gfx.hWnd,&bigrect);
    GetClientRect(gfx.hWnd,&smallrect);
    rightdiff = screen_width - smallrect.right;
    bottomdiff = screen_height - smallrect.bottom;
    if (gfx.hStatusBar)
    {
        GetClientRect(gfx.hStatusBar, &statusrect);
        bottomdiff += statusrect.bottom;
    }
    MoveWindow(gfx.hWnd, bigrect.left, bigrect.top, bigrect.right - bigrect.left + rightdiff, bigrect.bottom - bigrect.top + bottomdiff, TRUE);

    res = DirectDrawCreateEx(0, (LPVOID*)&lpdd, &IID_IDirectDraw7, 0);
    if (res != DD_OK)
    {
        DisplayError("Couldn't create a DirectDraw object.");
        return; /* to-do:  move to InitiateGFX? */
    }
    res = IDirectDraw_SetCooperativeLevel(lpdd, gfx.hWnd, DDSCL_NORMAL);
    if (res != DD_OK)
    {
        DisplayError("Couldn't set a cooperative level.");
        return; /* to-do:  move to InitiateGFX? */
    }

    zerobuf(&ddsd, sizeof(ddsd));
    ddsd.dwSize = sizeof(ddsd);
    ddsd.dwFlags = DDSD_CAPS;
    ddsd.ddsCaps.dwCaps = DDSCAPS_PRIMARYSURFACE;
    res = IDirectDraw_CreateSurface(lpdd, &ddsd, &lpddsprimary, 0);
    if (res != DD_OK)
    {
        DisplayError("CreateSurface for a primary surface failed.");
        return; /* to-do:  move to InitiateGFX? */
    }

    ddsd.dwFlags = DDSD_CAPS | DDSD_HEIGHT | DDSD_WIDTH | DDSD_PIXELFORMAT;
    ddsd.ddsCaps.dwCaps = DDSCAPS_OFFSCREENPLAIN;
    ddsd.dwWidth = PRESCALE_WIDTH;
    ddsd.dwHeight = PRESCALE_HEIGHT;
    zerobuf(&ftpixel, sizeof(ftpixel));
    ftpixel.dwSize = sizeof(ftpixel);
    ftpixel.dwFlags = DDPF_RGB;
    ftpixel.dwRGBBitCount = 32;
    ftpixel.dwRBitMask = 0xff0000;
    ftpixel.dwGBitMask = 0xff00;
    ftpixel.dwBBitMask = 0xff;
    ddsd.ddpfPixelFormat = ftpixel;
    res = IDirectDraw_CreateSurface(lpdd, &ddsd, &lpddsback, 0);
    if (res == DDERR_INVALIDPIXELFORMAT)
    {
        DisplayError(
            "ARGB8888 is not supported. You can try changing desktop color "\
            "depth to 32-bit, but most likely that won't help.");
        return; /* InitiateGFX fails. */
    }
    else if (res != DD_OK)
    {
        DisplayError("CreateSurface for a secondary surface failed.");
        return; /* InitiateGFX should fail. */
    }

    res = IDirectDrawSurface_GetSurfaceDesc(lpddsback, &ddsd);
    if (res != DD_OK)
    {
        DisplayError("GetSurfaceDesc failed.");
        return; /* InitiateGFX should fail. */
    }
    if ((ddsd.lPitch & 3) || ddsd.lPitch < (PRESCALE_WIDTH << 2))
    {
        DisplayError(
            "Pitch of a secondary surface is either not 32 bit aligned or "\
            "too small.");
        return; /* InitiateGFX should fail. */
    }
    pitchindwords = ddsd.lPitch >> 2;

    res = IDirectDraw_CreateClipper(lpdd, 0, &lpddcl, 0);
    if (res != DD_OK)
    {
        DisplayError("Couldn't create a clipper.");
        return; /* InitiateGFX should fail. */
    }
    res = IDirectDrawClipper_SetHWnd(lpddcl, 0, gfx.hWnd);
    if (res != DD_OK)
    {
        DisplayError("Couldn't register a windows handle as a clipper.");
        return; /* InitiateGFX should fail. */
    }
    res = IDirectDrawSurface_SetClipper(lpddsprimary, lpddcl);
    if (res != DD_OK)
    {
        DisplayError("Couldn't attach a clipper to a surface.");
        return; /* InitiateGFX should fail. */
    }

    src.top = src.left = 0; 
    src.bottom = 0;
#if SCREEN_WIDTH < PRESCALE_WIDTH
    src.right = PRESCALE_WIDTH - 1; /* fix for undefined video card behavior */
#else
    src.right = PRESCALE_WIDTH;
#endif
    p.x = p.y = 0;
    GetClientRect(gfx.hWnd, &dst);
    ClientToScreen(gfx.hWnd, &p);
    OffsetRect(&dst, p.x, p.y);
    GetClientRect(gfx.hStatusBar, &statusrect);
    dst.bottom -= statusrect.bottom;

    rdp_init();
 /* overlay = 1; */
    return;
}

EXPORT void CALL ShowCFB (void)
{
    MessageBox(NULL, "ShowCFB", NULL, MB_ICONWARNING);
    UpdateScreen();
    return;
}

EXPORT void CALL UpdateScreen(void)
{
    static int counter;

    if (counter++ < skip)
        return;
    counter = 0;
    rdp_update();
    if (step != 0)
        MessageBox(NULL, "Updated screen.\nPaused.", "Frame Step", MB_OK);
    return;
}

EXPORT void CALL ViStatusChanged (void)
{
}

EXPORT void CALL ViWidthChanged (void)
{
}

EXPORT void CALL FBWrite(DWORD addr, DWORD size)
{
    return;
}

EXPORT void CALL FBWList(FrameBufferModifyEntry *plist, DWORD size)
{
}

EXPORT void CALL FBRead(DWORD addr)
{
}

EXPORT void CALL FBGetFrameBufferInfo(void *pinfo)
{
}
